"""Fetch rates from openexchangerates with aiohttp."""

from .client import Client
from .exceptions import (
    OpenExchangeRatesAuthError,
    OpenExchangeRatesClientError,
    OpenExchangeRatesError,
)
from .model import Latest

__all__ = [
    "OpenExchangeRatesError",
    "OpenExchangeRatesAuthError",
    "OpenExchangeRatesClientError",
    "Client",
    "Latest",
]
__version__ = "0.4.14"



"""Provide common fixtures."""

from collections.abc import AsyncGenerator, Callable, Generator
from typing import Any

from aiohttp import ClientSession
from aioresponses import aioresponses
import pytest
from yarl import URL

from aioopenexchangerates.client import BASE_API_ENDPOINT, Client


@pytest.fixture(name="session")
async def session_fixture() -> AsyncGenerator[ClientSession, None]:
    """Provide a aiohttp client session."""
    async with ClientSession() as session:
        yield session


@pytest.fixture(name="client")
async def client_fixture(session: ClientSession) -> AsyncGenerator[Client, None]:
    """Provide a test client."""
    async with Client(
        api_key="test-api-key",
        session=session,
    ) as client:
        yield client


@pytest.fixture
def mock_response() -> Generator[aioresponses, None, None]:
    """Provide a mocker for aiohttp responses."""
    with aioresponses() as mock_response_:
        yield mock_response_


@pytest.fixture
def generate_url() -> Callable[..., str]:
    """Generate a URL from params."""

    def generate_url_(endpoint: str, **params: Any) -> str:
        return str(URL(f"{BASE_API_ENDPOINT}{endpoint}").with_query(params))

    return generate_url_



"""Provide exceptions for the Open Exchange Rates API."""


class OpenExchangeRatesError(Exception):
    """Represent an error for the Open Exchange Rates API."""


class OpenExchangeRatesClientError(OpenExchangeRatesError):
    """Represent a client error."""


class OpenExchangeRatesAuthError(OpenExchangeRatesClientError):
    """Represent an authentication error."""


class OpenExchangeRatesRateLimitError(OpenExchangeRatesClientError):
    """Represent a rate limit error."""



"""Test endpoint currencies.json."""

from collections.abc import Callable
import json

from aioresponses import aioresponses

from aioopenexchangerates.client import Client


async def test_get_latest_symbols(
    client: Client,
    mock_response: aioresponses,
    currencies: str,
    generate_url: Callable[..., str],
) -> None:
    """Test get_currencies."""
    mock_response.get(
        generate_url("currencies.json", show_alternative=0, show_inactive=0),
        body=currencies,
    )

    result = await client.get_currencies()

    assert result == json.loads(currencies)



"""Provide a base response model."""

from pydantic import BaseModel


class BaseResponse(BaseModel):
    """Represent a base response."""

    disclaimer: str
    license: str


class BaseRatesResponse(BaseResponse):
    """Represent a base rates response."""

    timestamp: int
    base: str
    rates: dict[str, float]



"""Provide a client for the Open Exchange Rates API."""

from __future__ import annotations

from http import HTTPStatus
from types import TracebackType
from typing import Any, cast

from aiohttp import ClientError, ClientResponse, ClientResponseError, ClientSession

from .exceptions import (
    OpenExchangeRatesAuthError,
    OpenExchangeRatesClientError,
    OpenExchangeRatesRateLimitError,
)
from .model import Latest

BASE_API_ENDPOINT = "https://openexchangerates.org/api/"


class Client:
    """Represent the client for the Open Exchange Rates API."""

    def __init__(self, api_key: str, session: ClientSession | None = None) -> None:
        """Initialize the client."""
        self.api_key = api_key
        self.session = session or ClientSession()

    async def request(self, endpoint: str, **kwargs: Any) -> ClientResponse:
        """Make a request."""
        url = f"{BASE_API_ENDPOINT}{endpoint}"
        try:
            return await self.session.get(url, raise_for_status=True, **kwargs)
        except ClientResponseError as err:
            if err.status == HTTPStatus.UNAUTHORIZED:
                raise OpenExchangeRatesAuthError(err.message) from err
            if err.status == HTTPStatus.FORBIDDEN:
                raise OpenExchangeRatesAuthError(err.message) from err
            if err.status == HTTPStatus.TOO_MANY_REQUESTS:
                raise OpenExchangeRatesRateLimitError(err.message) from err
            raise OpenExchangeRatesClientError(err.message) from err
        except ClientError as err:
            raise OpenExchangeRatesClientError(f"Unknown error: {err}") from err

    async def get_currencies(
        self, show_alternative: bool = False, show_inactive: bool = False
    ) -> dict[str, str]:
        """Get the supported currencies."""
        params = {
            "show_alternative": int(show_alternative),
            "show_inactive": int(show_inactive),
        }
        response = await self.request("currencies.json", params=params)
        return cast(dict[str, str], (await response.json()))

    async def get_latest(
        self, base: str = "USD", symbols: list[str] | None = None
    ) -> Latest:
        """Get the latest rates for the given base and symbols."""
        params = {"app_id": self.api_key, "base": base}
        if symbols:
            params["symbols"] = ",".join(symbols)
        response = await self.request("latest.json", params=params)
        return Latest(**(await response.json()))

    async def close(self) -> None:
        """Close the client."""
        await self.session.close()

    async def __aenter__(self) -> Client:
        """Enter the context manager."""
        return self

    async def __aexit__(
        self, exc_type: Exception, exc_value: str, traceback: TracebackType
    ) -> None:
        """Exit the context manager."""
        await self.close()



"""Configuration file for the Sphinx documentation builder."""

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from aioopenexchangerates import __version__

project = "aioopenexchangerates"
project_copyright = "2023, Martin Hjelmare"
author = "Martin Hjelmare"
version = __version__
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
]

# The suffix of source filenames.
source_suffix = [
    ".rst",
    ".md",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



"""Test the client."""

from collections.abc import Callable
from unittest.mock import patch

from aiohttp import ClientConnectionError, ClientPayloadError, ClientSession
from aioresponses import aioresponses
import pytest

from aioopenexchangerates.client import Client
from aioopenexchangerates.exceptions import (
    OpenExchangeRatesAuthError,
    OpenExchangeRatesClientError,
    OpenExchangeRatesRateLimitError,
)


@pytest.mark.parametrize(
    "status, error, message",
    [
        (401, OpenExchangeRatesAuthError, "Unauthorized"),
        (403, OpenExchangeRatesAuthError, "Forbidden"),
        (429, OpenExchangeRatesRateLimitError, "Too Many Requests"),
        (400, OpenExchangeRatesClientError, "Bad Request"),
        (500, OpenExchangeRatesClientError, "Internal Server Error"),
    ],
)
async def test_response_error(
    client: Client,
    mock_response: aioresponses,
    generate_url: Callable[..., str],
    status: int,
    error: type[Exception],
    message: str,
) -> None:
    """Test get_latest."""
    mock_response.get(
        generate_url("latest.json", app_id=client.api_key, base="USD"),
        status=status,
    )

    with pytest.raises(error) as err:
        await client.get_latest()

    assert str(err.value) == message


@pytest.mark.parametrize(
    "aiohttp_error, error, message",
    [
        (
            ClientConnectionError("Boom"),
            OpenExchangeRatesClientError,
            "Unknown error: Boom",
        ),
        (
            ClientPayloadError("Bad payload"),
            OpenExchangeRatesClientError,
            "Unknown error: Bad payload",
        ),
    ],
)
async def test_client_error(
    session: ClientSession,
    client: Client,
    aiohttp_error: Exception,
    error: type[Exception],
    message: str,
) -> None:
    """Test get_latest."""
    with patch.object(session, "get", side_effect=aiohttp_error), pytest.raises(
        error
    ) as err:
        await client.get_latest()

    assert str(err.value) == message



"""Provide model latest fixtures."""

import json
from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def latest_latest(latest_latest_usd: str, request: pytest.FixtureRequest) -> str:
    """Return a response for latest.json with USD base or request.param base."""
    body_string = latest_latest_usd
    body = json.loads(body_string)
    body["base"] = request.param or "USD"
    return json.dumps(body)


@pytest.fixture(scope="session")
def latest_latest_eur() -> str:
    """Return a response for latest.json with EUR base."""
    path = Path(__file__).parent.parent / "fixtures/latest/latest_eur.json"
    return path.read_text()


@pytest.fixture(name="latest_latest_usd", scope="session")
def latest_latest_usd_fixture() -> str:
    """Return a response for latest.json with USD base."""
    path = Path(__file__).parent.parent / "fixtures/latest/latest_usd.json"
    return path.read_text()


@pytest.fixture(name="latest_latest_usd_symbols", scope="session")
def latest_latest_usd_symbols_fixture() -> str:
    """Return a response for latest.json with USD base and symbols search."""
    path = Path(__file__).parent.parent / "fixtures/latest/latest_usd_symbols.json"
    return path.read_text()


@pytest.fixture(name="currencies", scope="session")
def currencies_fixture() -> str:
    """Return a response for currencies.json."""
    path = Path(__file__).parent.parent / "fixtures/currencies/currencies.json"
    return path.read_text()



"""Test endpoint latest.json."""

from collections.abc import Callable

from aioresponses import aioresponses
import pytest

from aioopenexchangerates.client import Client


@pytest.mark.parametrize(
    "base, latest_latest", [("USD", "USD"), ("EUR", "EUR")], indirect=["latest_latest"]
)
async def test_get_latest(
    client: Client,
    mock_response: aioresponses,
    latest_latest: str,
    generate_url: Callable[..., str],
    base: str,
) -> None:
    """Test get_latest."""
    mock_response.get(
        generate_url("latest.json", app_id=client.api_key, base=base),
        body=latest_latest,
    )

    result = await client.get_latest(base=base)

    assert result.disclaimer == "https://openexchangerates.org/terms/"
    assert result.license == "https://openexchangerates.org/license/"
    assert result.timestamp == 1449877801
    assert result.base == base
    assert result.rates == {
        "AED": 3.672538,
        "AFN": 66.809999,
        "ALL": 125.716501,
        "AMD": 484.902502,
        "ANG": 1.788575,
        "AOA": 135.295998,
        "ARS": 9.750101,
        "AUD": 1.390866,
    }


async def test_get_latest_symbols(
    client: Client,
    mock_response: aioresponses,
    latest_latest_usd_symbols: str,
    generate_url: Callable[..., str],
) -> None:
    """Test get_latest."""
    base = "USD"
    symbols = ["AMD", "ANG"]
    mock_response.get(
        generate_url(
            "latest.json", app_id=client.api_key, base=base, symbols=",".join(symbols)
        ),
        body=latest_latest_usd_symbols,
    )

    result = await client.get_latest(base=base, symbols=symbols)

    assert result.disclaimer == "https://openexchangerates.org/terms/"
    assert result.license == "https://openexchangerates.org/license/"
    assert result.timestamp == 1449877801
    assert result.base == base
    assert result.rates == {
        "AMD": 484.902502,
        "ANG": 1.788575,
    }
